/* ============================================================
 *
 * This file is part of the KDE project
 *
 * Date        : 2009-01-21
 * Description : Sane interface for KDE
 *
 * Copyright (C) 2009 by Kare Sars <kare dot sars at iki dot fi>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) version 3, or any
 * later version accepted by the membership of KDE e.V. (or its
 * successor approved by the membership of KDE e.V.), which shall
 * act as a proxy defined in Section 6 of version 3 of the license.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * ============================================================ */

#include "config.h"

// Local includes.
#include "sane_opt_combo.h"
#include "sane_opt_combo.moc"

// Qt includes
#include <QDebug>
#include <QtCore/QVarLengthArray>
#include <QIcon>
#include <QComboBox>

#include "../sane_i18n.h"

namespace LibQSane
{
static const char tmp_binary[] = "Binary";

SaneOptCombo::SaneOptCombo(const SANE_Handle handle, const int index)
: SaneOption(handle, index), m_combo(0)
{
}

QString SaneOptCombo::label()
{
    qDebug() << Q_FUNC_INFO << title();
    return title();
}

void SaneOptCombo::createWidget(QWidget *parent)
{
    if (m_widget) return;

    m_widget = m_combo = new QComboBox(parent);
    readOption();
    m_widget->setToolTip(description());
    connect(m_combo, SIGNAL(activated(int)), this, SLOT(comboboxChangedIndex(int)));
    readValue();
}

void SaneOptCombo::readValue()
{
    if (state() == STATE_HIDDEN) return;

    // read that current value
    QVarLengthArray<unsigned char> data(m_optDesc->size);
    SANE_Status status;
    SANE_Int res;
    status = sane_control_option (m_handle, m_index, SANE_ACTION_GET_VALUE, data.data(), &res);
    if (status != SANE_STATUS_GOOD) {
        return;
    }

    m_currentText = getSaneComboString(data.data());
    m_currentBAText = getSaneComboByteArray(data.data());
    if (m_combo != 0) {
        if (m_combo->currentText() != m_currentText) {
            setCurrentComboText(m_currentText);
            emit valueChanged();
        }
    }
}

void SaneOptCombo::readOption()
{
    SaneOption::readOption();
    
    if (!m_combo) return;
    
    QString saved = m_combo->currentText();
    
    QList<QPair<QString, QByteArray> > comboElements = genComboElements();
    
    m_combo->clear();
    
    for (int i = 0; i < comboElements.size(); i++)
        m_combo->addItem(comboElements.at(i).first, comboElements.at(i).second);
    
    setComboIcon(QIcon(ICONS_INSTALL_DIR"/color.png"), getSaneComboString((unsigned char*)SANE_VALUE_SCAN_MODE_COLOR));
    setComboIcon(QIcon(ICONS_INSTALL_DIR"/gray-scale.png"),
                     getSaneComboString((unsigned char*)SANE_VALUE_SCAN_MODE_GRAY));
    setComboIcon(QIcon(ICONS_INSTALL_DIR"/black-white.png"),
                     getSaneComboString((unsigned char*)SANE_VALUE_SCAN_MODE_LINEART));
                     // The epkowa/epson backend uses "Binary" which is the same as "Lineart"
    setComboIcon(QIcon(ICONS_INSTALL_DIR"/black-white.png"), tr(tmp_binary));
    
    // set the previous value
    setCurrentComboText(saved);
}

QList<QPair<QString, QByteArray> > SaneOptCombo::genComboElements()
{
    QList<QPair<QString, QByteArray> > result;
    
    QPair<QString, QByteArray> element;
    
    switch (m_optDesc->type)
    {
        case SANE_TYPE_INT:
            for (int i=1; i<=m_optDesc->constraint.word_list[0]; ++i)
            {
                element.first  = getSaneComboString((int)m_optDesc->constraint.word_list[i]);
                element.second = getSaneComboByteArray((int)m_optDesc->constraint.word_list[i]);
                result.append(element);
            }
            break;
        case SANE_TYPE_FIXED:
            for (int i=1; i<=m_optDesc->constraint.word_list[0]; ++i)
            {
                element.first  = getSaneComboString((float)SANE_UNFIX(m_optDesc->constraint.word_list[i]));
                element.second = getSaneComboByteArray((float)SANE_UNFIX(m_optDesc->constraint.word_list[i]));
                result.append(element);
            }
            break;
        case SANE_TYPE_STRING:
            {
                int i=0;
                while (m_optDesc->constraint.string_list[i] != 0)
                {
                    element.first  = getSaneComboString((unsigned char *)m_optDesc->constraint.string_list[i]);
                    element.second = getSaneComboByteArray((unsigned char *)m_optDesc->constraint.string_list[i]);
                    result.append(element);
                    i++;
                }
            }
            break;
        default :
            {
                element.first  = tr("NOT SUPPORTED");
                element.second = "NOT SUPPORTED";
                result.append(element);
            }
    }
    return result;
}

QString SaneOptCombo::getSaneComboString(int ival)
{
    switch(m_optDesc->unit)
    {
        case SANE_UNIT_NONE:        break;
        case SANE_UNIT_PIXEL:       return tr("%n Pixel(s)","", ival);
        case SANE_UNIT_BIT:         return tr("%n Bit(s)","", ival);
        case SANE_UNIT_MM:          return tr("%n mm","", ival);
        case SANE_UNIT_DPI:         return tr("%n DPI","", ival);
        case SANE_UNIT_PERCENT:     return tr("%n %","", ival);
        case SANE_UNIT_MICROSECOND: return tr("%n usec","", ival);
    }
    return QString::number(ival);
}

QByteArray SaneOptCombo::getSaneComboByteArray(int ival)
{
    QByteArray number = QByteArray::number(ival);
    switch(m_optDesc->unit)
    {
        case SANE_UNIT_NONE:        break;
        case SANE_UNIT_PIXEL:       return number + " Pixel(s)";
        case SANE_UNIT_BIT:         return number + " Bit(s)";
        case SANE_UNIT_MM:          return number + " mm";
        case SANE_UNIT_DPI:         return number + " DPI";
        case SANE_UNIT_PERCENT:     return number + " %";
        case SANE_UNIT_MICROSECOND: return number + " usec";
    }
    return number;
}

QString SaneOptCombo::getSaneComboString(float fval)
{
    switch(m_optDesc->unit)
    {
        case SANE_UNIT_NONE:        break;
        case SANE_UNIT_PIXEL:       return tr("%n Pixel(s)", "", fval);
        case SANE_UNIT_BIT:         return tr("%n Bit(s)","", fval);
        case SANE_UNIT_MM:          return tr("%n mm", "", fval);
        case SANE_UNIT_DPI:         return tr("%n DPI", "", fval);
        case SANE_UNIT_PERCENT:     return tr("%n %", "", fval);
        case SANE_UNIT_MICROSECOND: return tr("%n usec","", fval);
    }
    return QString::number(fval, 'F', 4);
}

QByteArray SaneOptCombo::getSaneComboByteArray(float fval)
{
    QByteArray number = QByteArray::number(fval, 'F', 4);
    switch(m_optDesc->unit)
    {
        case SANE_UNIT_NONE:        break;
        case SANE_UNIT_PIXEL:       return number + " Pixel(s)";
        case SANE_UNIT_BIT:         return number + " Bit(s)";
        case SANE_UNIT_MM:          return number + " mm";
        case SANE_UNIT_DPI:         return number + " DPI";
        case SANE_UNIT_PERCENT:     return number + " %";
        case SANE_UNIT_MICROSECOND: return number + " usec";
    }
    return number;
}

QString SaneOptCombo::getSaneComboString(unsigned char *data)
{
    QString tmp;
    if (data == 0) return QString();
        
    switch (m_optDesc->type)
    {
        case SANE_TYPE_INT:
            return getSaneComboString((int)toSANE_Word(data));
        case SANE_TYPE_FIXED:
            return getSaneComboString((float)SANE_UNFIX(toSANE_Word(data)));
        case SANE_TYPE_STRING:
            tmp = sane_i18n(reinterpret_cast<char*>(data));
            tmp = tmp.simplified();
            if (tmp.length() > 25) {
                tmp = tmp.left(22);
                tmp += "...";
            }
            return tmp;
        default :
            break;
    }
    return QString();
}

QByteArray SaneOptCombo::getSaneComboByteArray(unsigned char *data)
{
    QByteArray tmp;
    if (data == 0) return QByteArray();
        
    switch (m_optDesc->type)
    {
        case SANE_TYPE_INT:
            return getSaneComboByteArray((int)toSANE_Word(data));
        case SANE_TYPE_FIXED:
            return getSaneComboByteArray((float)SANE_UNFIX(toSANE_Word(data)));
        case SANE_TYPE_STRING:
            tmp = QByteArray(reinterpret_cast<char*>(data));
            tmp = tmp.simplified();
            if (tmp.length() > 25) {
                tmp = tmp.left(22);
                tmp += "...";
            }
            return tmp;
        default :
            break;
    }
    return QByteArray();
}

void SaneOptCombo::comboboxChangedIndex(int i)
{
    if (m_combo && (m_combo->currentText() == m_currentText)) {
        return;
    }
    
    unsigned char data[4];
    void *dataPtr;
    
    switch (m_optDesc->type)
    {
        case SANE_TYPE_INT:
        case SANE_TYPE_FIXED:
            fromSANE_Word(data, m_optDesc->constraint.word_list[i+1]);
            dataPtr = data;
            break;
        case SANE_TYPE_STRING:
            dataPtr = (void *)m_optDesc->constraint.string_list[i];
            break;
        default:
            qDebug() << "can not handle type:" << m_optDesc->type;
            return;
    }
    writeData(dataPtr);
    readValue();
    emit valueChanged();
}

bool SaneOptCombo::getMinValue(float &val)
{
    if (state() == STATE_HIDDEN) return false;
    switch (m_optDesc->type)
    {
        case SANE_TYPE_INT:
            val = (float)m_optDesc->constraint.word_list[1];
            for (int i=2; i<=m_optDesc->constraint.word_list[0]; i++) {
                val = qMin((float)m_optDesc->constraint.word_list[i], val);
            }
            break;
        case SANE_TYPE_FIXED:
            val = (float)SANE_UNFIX(m_optDesc->constraint.word_list[1]);
            for (int i=2; i<=m_optDesc->constraint.word_list[0]; i++) {
                val = qMin((float)SANE_UNFIX(m_optDesc->constraint.word_list[i]), val);
            }
            break;
        default:
            qDebug() << "can not handle type:" << m_optDesc->type;
            return false;
    }
    return true;
}

bool SaneOptCombo::getValue(float &val)
{
    if (state() == STATE_HIDDEN) return false;

    // read that current value
    QVarLengthArray<unsigned char> data(m_optDesc->size);
    SANE_Status status;
    SANE_Int res;
    status = sane_control_option (m_handle, m_index, SANE_ACTION_GET_VALUE, data.data(), &res);
    if (status != SANE_STATUS_GOOD) {
        qDebug() << m_optDesc->name << "sane_control_option returned" << status;
        return false;
    }
    
    switch (m_optDesc->type)
    {
        case SANE_TYPE_INT:
            val = (float)toSANE_Word(data.data());
            return true;
        case SANE_TYPE_FIXED:
            val = SANE_UNFIX(toSANE_Word(data.data()));
            return true;
        default:
            qDebug() << "Type" << m_optDesc->type << "not supported!";
    }
    return false;
}

bool SaneOptCombo::setValue(float value)
{
    unsigned char data[4];
    float tmp;
    float minDiff;
    int i;
    int minIndex = 1;
    
    switch (m_optDesc->type)
    {
        case SANE_TYPE_INT:
            tmp = (float)m_optDesc->constraint.word_list[minIndex];
            minDiff = qAbs(value - tmp);
            for (i=2; i<=m_optDesc->constraint.word_list[0]; ++i) {
                tmp = (float)m_optDesc->constraint.word_list[i];
                if (qAbs(value - tmp) < minDiff) {
                    minDiff = qAbs(value - tmp);
                    minIndex = i;
                }
            }
            fromSANE_Word(data, m_optDesc->constraint.word_list[minIndex]);
            writeData(data);
            readValue();
            return (minDiff < 1.0);
        case SANE_TYPE_FIXED:
            tmp = (float)SANE_UNFIX(m_optDesc->constraint.word_list[minIndex]);
            minDiff = qAbs(value - tmp);
            for (i=2; i<=m_optDesc->constraint.word_list[0]; ++i) {
                tmp = (float)SANE_UNFIX(m_optDesc->constraint.word_list[i]);
                if (qAbs(value - tmp) < minDiff) {
                    minDiff = qAbs(value - tmp);
                    minIndex = i;
                }
            }
            fromSANE_Word(data, m_optDesc->constraint.word_list[minIndex]);
            writeData(data);
            readValue();
            return (minDiff < 1.0);
        default:
            qDebug() << "can not handle type:" << m_optDesc->type;
            break;
    }
    return false;
}

bool SaneOptCombo::getValue(QString &val)
{
    if (state() == STATE_HIDDEN) return false;
    val = m_currentText;
    return true;
}

bool SaneOptCombo::setValue(const QString &val)
{
    if (state() == STATE_HIDDEN) return false;
    if (val == m_currentText) return true;
    
    unsigned char data[4];
    void *data_ptr;
    SANE_Word fixed;
    int i;
    float f;
    bool ok;
    QString tmp;
    
    switch (m_optDesc->type)
    {
        case SANE_TYPE_INT:
            tmp = val.left(val.indexOf(' ')); // strip the unit
            // accept float formating of the string
            i = (int)(tmp.toFloat(&ok));
            if (ok == false) return false;
            fromSANE_Word(data, i);
            data_ptr = data;
            break;
        case SANE_TYPE_FIXED:
            tmp = val.left(val.indexOf(' ')); // strip the unit
            f = tmp.toFloat(&ok);
            if (ok == false) return false;
            fixed = SANE_FIX(f);
            fromSANE_Word(data, fixed);
            data_ptr = data;
            break;
        case SANE_TYPE_STRING:
            i = 0;
            while (m_optDesc->constraint.string_list[i] != 0) {
                tmp = getSaneComboString((unsigned char *)m_optDesc->constraint.string_list[i]);
                if (val == tmp) {
                    data_ptr = (void *)m_optDesc->constraint.string_list[i];
                    break;
                }
                i++;
            }
            if (m_optDesc->constraint.string_list[i] == 0) return false;
            break;
        default:
            qDebug() << "can only handle SANE_TYPE: INT, FIXED and STRING";
            return false;
    }
    writeData(data_ptr);

    readValue();
    return true;
}

bool SaneOptCombo::getRawValue(QByteArray &val)
{
    if (state() == STATE_HIDDEN) return false;
    val = m_currentBAText;
    return true;
}

bool SaneOptCombo::setRawValue(const QByteArray &val)
{
    if (state() == STATE_HIDDEN) return false;
    if (val == m_currentBAText) return true;
    
    unsigned char data[4];
    void *data_ptr;
    SANE_Word fixed;
    int i;
    float f;
    bool ok;
    QByteArray tmp;
    
    switch (m_optDesc->type)
    {
        case SANE_TYPE_INT:
            tmp = val.left(val.indexOf(' ')); // strip the unit
            // accept float formating of the string
            i = (int)(tmp.toFloat(&ok));
            if (ok == false) return false;
            fromSANE_Word(data, i);
            data_ptr = data;
            break;
        case SANE_TYPE_FIXED:
            tmp = val.left(val.indexOf(' ')); // strip the unit
            f = tmp.toFloat(&ok);
            if (ok == false) return false;
            fixed = SANE_FIX(f);
            fromSANE_Word(data, fixed);
            data_ptr = data;
            break;
        case SANE_TYPE_STRING:
            i = 0;
            while (m_optDesc->constraint.string_list[i] != 0) {
                tmp = getSaneComboByteArray((unsigned char *)m_optDesc->constraint.string_list[i]);
                if (val == tmp) {
                    data_ptr = (void *)m_optDesc->constraint.string_list[i];
                    break;
                }
                i++;
            }
            if (m_optDesc->constraint.string_list[i] == 0) return false;
            break;
        default:
            qDebug() << "can only handle SANE_TYPE: INT, FIXED and STRING";
            return false;
    }
    writeData(data_ptr);

    readValue();
    return true;
}

void SaneOptCombo::setCurrentComboText(const QString& t)
{
    int i;
    for (i=0; i<m_combo->count(); i++) {
        if (m_combo->itemText(i) == t) {
            m_combo->setCurrentIndex(i);
        }
    }
}

bool SaneOptCombo::setComboIcon(const QIcon& icon, const QString& str)
{
    for (int i=0; i < m_combo->count(); i++) {
        if (m_combo->itemText(i) == str) {
            m_combo->setItemIcon(i, icon);
            return true;
        }
    }
    return false;
}

}  // NameSpace LibQSane
