/*
    <one line to give the library's name and an idea of what it does.>
    Copyright (C) <year>  <name of author>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

*/

#include "previewimagechunkupdater.h"

#include <QDebug>
#include <QImage>
#include <QPainter>

namespace LibQSane
{

PreviewImageChunkUpdater::PreviewImageChunkUpdater(QImage* img)
    : m_img(img),
      m_frameRead(0),
      m_pixel_x(0),
      m_pixel_y(0),
      m_pixel_idx(0),
      m_px_c_index(0),
      m_chunkIndex(0)
{
    m_px_colors[0] = 0;
    m_px_colors[1] = 0;
    m_px_colors[2] = 0;
}

void PreviewImageChunkUpdater::start(const SANE_Parameters& params, const QRect& previewRect, float scanDPI)
{
    qDebug() << Q_FUNC_INFO;
    
    beginFrame(params);
    
    m_previewRect = previewRect;
    
    m_chunkImage = QImage(params.pixels_per_line, qRound(scanDPI / 25.0), QImage::Format_RGB32);
    m_chunkImage.fill(0xFFFFFFFF);   
}

void PreviewImageChunkUpdater::beginFrame(const SANE_Parameters& params)
{
    qDebug() << Q_FUNC_INFO;
    
    m_params = params;
    
    m_frameRead   = 0;
    
    m_pixel_x     = 0;
    m_pixel_y     = 0;
    m_pixel_idx   = 0;
    
    m_px_c_index  = 0;
    
    m_chunkIndex = 0;
}

#define inc_pixel(x,y,ppl) { x++; if (x>=ppl) { y++; x=0;} }
#define inc_color_index(index) { index++; if (index==3) index=0;}

bool PreviewImageChunkUpdater::update(const SANE_Byte readData[], int readBytes)
{
    if (m_chunkImage.isNull())
        return false;
    
    uchar *imgBits = m_chunkImage.bits();
    switch (m_params.format)
    {
        case SANE_FRAME_GRAY:
            if (m_params.depth == 1) {
                for (int i=0; i<readBytes; i++) {
                    for (int j=7; j>=0; --j)
                    {
                        QRgb rgb = (readData[i] & (1<<j)) == 0 ? qRgb(255,255,255) : qRgb(0,0,0);
                        m_chunkImage.setPixel(m_pixel_x, m_pixel_y, rgb);
                        
                        m_pixel_x++;
                        if (m_pixel_x >= m_chunkImage.width()) {
                            m_pixel_x = 0;
                            m_pixel_y++;
                            if (m_pixel_y >= m_chunkImage.height())
                                flushChunk();
                            
                            break;
                        }
                    }
                    m_frameRead++;
                }
                return true;
            }
            else if (m_params.depth == 8) {
                int flushIdx = m_chunkImage.width() * m_chunkImage.height();
                for (int i=0; i<readBytes; i++) {
                    int index = m_pixel_idx * 4;
                    imgBits[index    ] = readData[i];
                    imgBits[index + 1] = readData[i];
                    imgBits[index + 2] = readData[i];
                    m_pixel_idx ++;
                    m_frameRead++;
                    
                    if (m_pixel_idx >= flushIdx)
                        flushChunk();
                }
                return true;
            }
            else if (m_params.depth == 16) {
                int flushIdx = m_chunkImage.width() * m_chunkImage.height();
                for (int i=0; i<readBytes; i++) {
                    if (m_frameRead%2 == 1) {
                        int index = m_pixel_idx * 4;
                        imgBits[index    ] = readData[i]; // little-endian
                        imgBits[index + 1] = readData[i];
                        imgBits[index + 2] = readData[i];
                        m_pixel_idx++;
                        
                        if (m_pixel_idx == flushIdx)
                            flushChunk();
                    }
                    m_frameRead++;
                }
                return true;
            }
            break;
            
        case SANE_FRAME_RGB:
            if (m_params.depth == 8) {
                for (int i=0; i<readBytes; i++) {
                    m_px_colors[m_px_c_index] = readData[i];
                    inc_color_index(m_px_c_index);
                    
                    if (m_px_c_index == 0) {
                        m_chunkImage.setPixel(m_pixel_x,
                                        m_pixel_y,
                                        qRgb(m_px_colors[0],
                                                m_px_colors[1],
                                                m_px_colors[2]));
                        inc_pixel(m_pixel_x, m_pixel_y, m_chunkImage.width());
                        
                        if (m_pixel_y >= m_chunkImage.height())
                            flushChunk();
                    }
                    
                    m_frameRead++;
                }
                return true;
            }
            else if (m_params.depth == 16) {
                for (int i=0; i<readBytes; i++) {
                    if (m_frameRead%2==1) {
                        m_px_colors[m_px_c_index] = readData[i];
                        inc_color_index(m_px_c_index);
                        if (m_px_c_index == 0) {
                            m_chunkImage.setPixel(m_pixel_x,
                                            m_pixel_y,
                                            qRgb(m_px_colors[0],
                                                    m_px_colors[1],
                                                    m_px_colors[2]));
                            inc_pixel(m_pixel_x, m_pixel_y, m_params.pixels_per_line);
                            
                            if (m_pixel_y >= m_chunkImage.height())
                                flushChunk();
                        }
                    }
                    m_frameRead++;
                }
                return true;
            }
            break;
        
        case SANE_FRAME_RED:
        case SANE_FRAME_GREEN:
        case SANE_FRAME_BLUE:
            if (m_params.depth == 8) {
                int flushIdx = m_chunkImage.width() * m_chunkImage.height();
                for (int i=0; i<readBytes; i++) {
                    imgBits[m_pixel_idx * 4] = readData[i];
                    m_frameRead++;
                    m_pixel_idx++;
                    
                    if (m_pixel_idx >= flushIdx)
                        flushOneColorChunk();
                }
                return true;
            }
            else if (m_params.depth == 16) {
                int flushIdx = m_chunkImage.width() * m_chunkImage.height();
                for (int i=0; i<readBytes; i++) {
                    if (m_frameRead%2 == 1) {
                        imgBits[m_pixel_idx * 4] = readData[i];
                        m_pixel_idx++;
                        
                        if (m_pixel_idx >= flushIdx)
                            flushOneColorChunk();
                    }
                    m_frameRead++;
                }
                return true;
            }
            break;
    }

    qWarning() << "Format" << m_params.format
    << "and depth" << m_params.format
    << "is not yet supported by libqsane!";
    
    return false;
}

void PreviewImageChunkUpdater::flushChunk()
{
    m_pixel_x = 0;
    m_pixel_y = 0;
    m_pixel_idx = 0;
    m_chunkIndex++;
    
    double scale = (double)m_previewRect.width() / (double)m_chunkImage.width();
    int relativeYPosTop    = qRound(scale * m_chunkIndex * m_chunkImage.height());
    int relativeYPosBottom = qRound(scale * (m_chunkIndex+1) * m_chunkImage.height());
    
    QRect previewRect(m_previewRect.x(), m_previewRect.y() + relativeYPosTop, m_previewRect.width(), relativeYPosBottom - relativeYPosTop);
    
    QPainter p(m_img);
    p.setRenderHint(QPainter::SmoothPixmapTransform);
    p.drawImage(previewRect, m_chunkImage);
}

void PreviewImageChunkUpdater::flushOneColorChunk()
{
    m_pixel_idx = 0;
    m_chunkIndex++;
    
    double scale = (double)m_previewRect.width() / (double)m_chunkImage.width();
    int relativeYPosTop    = qRound(scale * m_chunkIndex * m_chunkImage.height());
    int relativeYPosBottom = qRound(scale * (m_chunkIndex+1) * m_chunkImage.height());
    
    QImage previewChunk = m_chunkImage.scaled(m_previewRect.width(), relativeYPosBottom - relativeYPosTop,
                                              Qt::IgnoreAspectRatio, Qt::SmoothTransformation);

    if (previewChunk.isNull())
        return;
    
    uchar* chunkBits = previewChunk.bits();
    
    uchar* imgBits = m_img->bits() + 4*m_previewRect.x() + (m_previewRect.y() + relativeYPosTop)*m_img->bytesPerLine();
    uchar* maxImgBits = m_img->bits() + m_img->byteCount() - 1;
    
    if (imgBits > maxImgBits)
        return;
    
    int colorOffset = 0;
    switch (m_params.format)
    {
        case SANE_FRAME_RED:
            colorOffset = 2;
            break;
        case SANE_FRAME_GREEN:
            colorOffset = 1;
            break;
        case SANE_FRAME_BLUE:
            colorOffset = 0;
            break;
        default:
            return;
    }
    
    for (int y = 0; y < previewChunk.height(); y++)
    {
        for (int x = 0; x < previewChunk.width(); x++)
        {
            if (imgBits + x*4 + colorOffset > maxImgBits)
                break;
            
            imgBits[x*4 + colorOffset] = chunkBits[x*4];
        }
        
        chunkBits += previewChunk.bytesPerLine();
        imgBits += m_img->bytesPerLine();
        if (imgBits > maxImgBits)
            break;
    }
}

}
