/* ============================================================
 *
 * This file is part of the KDE project
 *
 * Date        : 2007-09-13
 * Description : Sane interface for KDE
 *
 * Copyright (C) 2007-2008 by Kare Sars <kare dot sars at iki dot fi>
 * Copyright (C) 2009 by Grzegorz Kurtyka <grzegorz dot kurtyka at gmail dot com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) version 3, or any
 * later version accepted by the membership of KDE e.V. (or its
 * successor approved by the membership of KDE e.V.), which shall
 * act as a proxy defined in Section 6 of version 3 of the license.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * ============================================================ */

// Local includes.
#include "sane_device_dialog.h"
#include "sane_device_dialog_p.h"
#include "moc_sane_device_dialog.cpp"

// Sane includes.
extern "C"
{
#include <sane/saneopts.h>
#include <sane/sane.h>
}


// Qt includes.
#include <QDebug>

#include <QPushButton>
#include <QDialogButtonBox>
#include <QScrollArea>
#include <QGroupBox>
#include <QStringList>
#include <QButtonGroup>
#include <QToolTip>
#include <QRadioButton>
#include <QCheckBox>
#include <QPushButton>
#include <QBoxLayout>

#include "sane_find_devices_thread.h"

namespace LibQSane
{

SaneDeviceDialog::SaneDeviceDialog(QWidget *parent, bool autoSelectSingleDevice)
    : QDialog(parent),
      d_ptr(new SaneDeviceDialogPrivate())
{
    Q_D(SaneDeviceDialog);
    d->q_ptr = this;
    d->m_autoSelectSingleDevice = autoSelectSingleDevice;

    d->init();
}

SaneDeviceDialog::~SaneDeviceDialog() {
    ///@todo wait for thread to finish if its running
}

void SaneDeviceDialog::setDefault(QString defaultDevice)
{
    Q_D(SaneDeviceDialog);
    d->m_selectedDevice = defaultDevice;
}

QString SaneDeviceDialog::getSelectedName()
{
    Q_D(SaneDeviceDialog);
    QAbstractButton *selected_button = d->m_btnGroup->checkedButton();
    if(selected_button) {
        return selected_button->objectName();
    }
    return QString();
}

SaneDeviceInfo SaneDeviceDialog::getSelectedDevice()
{
    Q_D(SaneDeviceDialog);
    QAbstractButton *selectedButton = d->m_btnGroup->checkedButton();
    if (selectedButton)
    {
        QString deviceName = selectedButton->objectName();
        return d->m_deviceInfosMap.value(deviceName);
    }
    
    return SaneDeviceInfo();
}

void SaneDeviceDialog::reloadDevicesList()
{
    Q_D(SaneDeviceDialog);
    if(!d->m_findDevThread->isRunning()) {
        d->_q_setAvailable(false);

        while (!d->m_btnGroup->buttons().isEmpty()) {
            delete d->m_btnGroup->buttons().takeFirst();
        }

        d->m_btnBox->setEnabled(false);
        d->m_btnBox->setTitle( tr("Looking for devices. Please wait.") );
        d->m_reloadButton->setEnabled(false);
        d->m_findDevThread->start();
    }
}

//////////////////////////////////////////////////

SaneDeviceDialogPrivate::SaneDeviceDialogPrivate()
{
}

SaneDeviceDialogPrivate::~SaneDeviceDialogPrivate()
{
}

void SaneDeviceDialogPrivate::init()
{
    Q_Q(SaneDeviceDialog);
    
    q->setWindowTitle(SaneDeviceDialog::tr("Select device"));
    
    m_buttonBox = new QDialogButtonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel);
    
    m_reloadButton = new QPushButton(SaneDeviceDialog::tr("Reload devices list"));
    
    //m_buttonBox->addButton(m_reloadButton, QDialogButtonBox::ActionRole);
    QHBoxLayout* buttonsLayout = new QHBoxLayout;
    buttonsLayout->addWidget(m_reloadButton);
    buttonsLayout->addWidget(m_buttonBox);
    
    QObject::connect(m_buttonBox, SIGNAL(accepted()), q, SLOT(accept()));
    QObject::connect(m_buttonBox, SIGNAL(rejected()), q, SLOT(reject()));
    
    m_btnGroup = new QButtonGroup(q);
    
    m_btnBox = new QGroupBox;
    QVBoxLayout *layout = new QVBoxLayout;
    m_btnContainer = new QWidget;
    m_btnLayout = new QVBoxLayout(m_btnContainer);
    QScrollArea *area = new QScrollArea;
    
    m_btnBox->setLayout(layout);
    
    layout->addWidget(area);
    layout->setContentsMargins(0,0,0,0);

    area->setWidgetResizable(true);
    area->setFrameShape(QFrame::NoFrame);
    area->setWidget(m_btnContainer);
    
    QVBoxLayout* mainLayout = new QVBoxLayout;
    
    mainLayout->addWidget(m_btnBox);
    mainLayout->addLayout(buttonsLayout);
    
    q->setLayout(mainLayout);
    
    m_findDevThread = new FindSaneDevicesThread(q);

    QObject::connect(m_findDevThread, SIGNAL( finished() ),
            q, SLOT( _q_updateDevicesList() ) );
    
    QObject::connect(m_reloadButton, SIGNAL(clicked()),
             q, SLOT(reloadDevicesList()) );
    
    q->setMinimumHeight(200);
    q->resize(500, 200);
    _q_setAvailable(false);
    q->reloadDevicesList();
}

void SaneDeviceDialogPrivate::_q_updateDevicesList()
{
    QMap<QString,SaneDeviceInfo> devicesList;
    m_findDevThread->getDevicesList(devicesList);
    setDevicesList( devicesList );
    m_btnBox->setEnabled(true);
    m_reloadButton->setEnabled(true);
}

void SaneDeviceDialogPrivate::_q_setAvailable(bool avail)
{
    Q_Q(SaneDeviceDialog);
    m_buttonBox->button(QDialogButtonBox::Ok)->setEnabled(avail);
    
    if(avail) {
        m_selectedDevice = q->getSelectedName();
        m_buttonBox->button(QDialogButtonBox::Ok)->setFocus();
    }
}

bool SaneDeviceDialogPrivate::setDevicesList(const QMap<QString, SaneDeviceInfo>& items)
{
    Q_Q(SaneDeviceDialog);
    
    m_deviceInfosMap = items;
    
    while (!m_btnGroup->buttons().isEmpty()) {
        delete m_btnGroup->buttons().takeFirst();
    }

    if (items.isEmpty()) {
        m_btnBox->setTitle( SaneDeviceDialog::tr("Sorry. No devices found.") );
        return false;
    }

    delete m_btnLayout;
    m_btnLayout = new QVBoxLayout;
    m_btnContainer->setLayout(m_btnLayout);
    m_btnBox->setTitle( SaneDeviceDialog::tr("Found devices:") );
    QMapIterator<QString, SaneDeviceInfo> itr(items);
    while (itr.hasNext()) {
        bool first = !itr.hasPrevious();
        itr.next();
        SaneDeviceInfo deviceInfo = itr.value();
        QString rbLabel = deviceInfo.vendor;
        rbLabel += " : " + deviceInfo.model;
        rbLabel += "\n " + deviceInfo.name;
        QRadioButton *b = new QRadioButton(rbLabel, q );
        b->setObjectName(itr.key());
        b->setToolTip( itr.key() );
        m_btnLayout->addWidget(b);
        m_btnGroup->addButton(b);
        q->connect(b, SIGNAL(clicked(bool)), q, SLOT(_q_setAvailable(bool)) );
        if(first || itr.key() == m_selectedDevice) {
            b->setChecked(true);
            _q_setAvailable(true);
        }
    }

    m_btnLayout->addStretch();

    if (m_autoSelectSingleDevice && items.size() == 1) {
        m_buttonBox->button(QDialogButtonBox::Ok)->animateClick();
    }

    return true;
}

}  // NameSpace LibQSane
