#include "sane_image_data.h"

#include <QCoreApplication>
#include <QSharedData>

namespace LibQSane
{

class SaneImageDataPrivate : public QSharedData
{
public:
    QByteArray m_data;
    int m_width;
    int m_height;
    int m_bytesPerLine;
    SaneImageData::ImageFormat m_format;
    int m_dpi;
};

SaneImageData::SaneImageData()
    : d(new SaneImageDataPrivate)
{
    d->m_width = 0;
    d->m_height = 0;
    d->m_bytesPerLine = 0;
    d->m_format = FormatNone;
    d->m_dpi = 0;
}

SaneImageData::SaneImageData(QByteArray data, int width, int height, int bytesPerLine, ImageFormat format, int dpi)
    : d(new SaneImageDataPrivate)
{
    d->m_data = data;
    d->m_width = width;
    d->m_height = height;
    d->m_bytesPerLine = bytesPerLine;
    d->m_format = format;
    d->m_dpi = dpi;
}

SaneImageData::SaneImageData(const SaneImageData& other)
    : d(other.d)
{
}

SaneImageData& SaneImageData::operator=(const SaneImageData& other)
{
    d = other.d;
    return *this;
}

SaneImageData::~SaneImageData()
{
}

QByteArray SaneImageData::data() const
{
    return d->m_data;
}

int SaneImageData::width() const
{
    return d->m_width;
}

int SaneImageData::height() const
{
    return d->m_height;
}

int SaneImageData::bytesPerLine() const
{
    return d->m_bytesPerLine;
}

SaneImageData::ImageFormat SaneImageData::format() const
{
    return d->m_format;
}

int SaneImageData::dpi() const
{
    return d->m_dpi;
}

#define inc_pixel(x,y,ppl) { x++; if (x>=ppl) { y++; x=0;} }

QImage SaneImageData::toQImage(QString* error, QString* warning) const
{
    QImage img;
    int j=0;
    int pixel_x = 0;
    int pixel_y = 0;
    QVector<QRgb> table;
    
    QByteArray data = d->m_data;
    int width = d->m_width;
    int height = d->m_height;
    int bytes_per_line = d->m_bytesPerLine;
    ImageFormat format = d->m_format;

    if (error != 0)
        error->clear();
    
    if (warning != 0)
        warning->clear();
    
    switch (format)
    {
        case FormatBlackWhite:
            img = QImage((uchar*)data.data(),
                          width,
                          height,
                          bytes_per_line,
                          QImage::Format_Mono);
            // The color table must be set
            table.append(0xFFFFFFFF);
            table.append(0xFF000000);
            img.setColorTable(table);
            break;

        case FormatGrayScale8:
            img = QImage(width,
                         height,
                         QImage::Format_RGB32);
            j=0;
            for (int i=0; i<data.size(); i++) {
                img.bits()[j+0] = data.data()[i];
                img.bits()[j+1] = data.data()[i];
                img.bits()[j+2] = data.data()[i];
                j+=4;
            }
            break;

        case FormatGrayScale16:
            img = QImage(width,
                         height,
                         QImage::Format_RGB32);
            j=0;
            for (int i=1; i<data.size(); i+=2) {
                img.bits()[j+0] = data.data()[i];
                img.bits()[j+1] = data.data()[i];
                img.bits()[j+2] = data.data()[i];
                j+=4;
            }
            
            if (warning != 0)
                *warning = qApp->translate("SaneImageData", "The image data contained 16 bits per color, "
                    "but the color depth has been truncated to 8 bits per color.");
            break;

        case FormatRGB_8_C:
            pixel_x = 0;
            pixel_y = 0;

            img = QImage(width,
                         height,
                         QImage::Format_RGB32);

            for (int i=0; i<data.size(); i+=3) {
                img.setPixel(pixel_x,
                             pixel_y,
                             qRgb(data[i],
                                  data[i+1],
                                  data[i+2]));

                inc_pixel(pixel_x, pixel_y, width);
            }
            break;

        case FormatRGB_16_C:
            pixel_x = 0;
            pixel_y = 0;

            img = QImage(width,
                         height,
                         QImage::Format_RGB32);

            for (int i=1; (i+4) < data.size(); i+=6) {
                img.setPixel(pixel_x,
                             pixel_y,
                             qRgb(data[i],
                                  data[i+2],
                                  data[i+4]));

                inc_pixel(pixel_x, pixel_y, width);
            }
            if (warning != 0)
                *warning = qApp->translate("SaneImageData", "The image data contained 16 bits per color, "
                    "but the color depth has been truncated to 8 bits per color.");
            break;

        case FormatNone:
        default:
            if (error != 0)
                *error = qApp->translate("SaneImageData", "The image data is of unsupported format.");
            break;
    }
    
    if (img.isNull())
    {
        if (error != 0 && error->isNull())
            *error = qApp->translate("SaneImageData", "Unknown error");
    }
    else
    {
        float dpm = d->m_dpi * (1000.0 / 25.4);
        img.setDotsPerMeterX(dpm);
        img.setDotsPerMeterY(dpm);
    }
    
    return img;
}

}
