#ifndef SANE_IMAGE_DATA_H
#define SANE_IMAGE_DATA_H

#include <QByteArray>
#include <QImage>
#include <QSharedDataPointer>

#include "libqsane_export.h"

namespace LibQSane
{

class SaneImageDataPrivate;

class LIBQSANE_EXPORT SaneImageData
{
public:
    /** This enumeration describes the type of the returned data.
     * The number of formats might grow, so it is wise to be prepard fro more.*/
    typedef enum
    {
        FormatBlackWhite,   /**< One bit per pixel 1 = black 0 = white */
        FormatGrayScale8,   /**< Grayscale with one byte per pixel 0 = black 255 = white */
        FormatGrayScale16,  /**< Grayscale withtTwo bytes per pixel.
                             * The byte order is the one provided by libsane. */
        FormatRGB_8_C,      /**< Every pixel consists of three colors in the order Read,
                             * Grean and Blue, with one byte per color (no alpha channel). */
        FormatRGB_16_C,     /**< Every pixel consists of three colors in the order Read,
                             * Grean and Blue, with two bytes per color(no alpha channel).
                             * The byte order is the one provided by libsane. */
        FormatBMP,          /**< The image data  is returned as a BMP. */
        FormatNone = 0xFFFF /**< This enumeration value should never be returned to the user */
    } ImageFormat;
    
    SaneImageData();
    SaneImageData(QByteArray data, int width, int height, int bytesPerLine, ImageFormat format, int dpi);
    SaneImageData(const SaneImageData& other);
    SaneImageData& operator=(const SaneImageData& other);
    virtual ~SaneImageData();
    
    /**
     * @return This function returns the byte data containing the image. */
    QByteArray data() const;
    
    /**
     * @return This function returns the width of the image in pixels. */
    int width() const;
    
    /**
     * @return This function returns the height of the image in pixels. */
    int height() const;
    
    /**
     * @return This function returns the number of bytes used per line.
     * This might include padding and is probably only relevant for 'FormatBlackWhite'. */
    int bytesPerLine() const;
    
    /**
     * @return This function returns the Sane image format of the data. */
    ImageFormat format() const;
    
    /**
     * @return This function returns the resolution of the image in dots per inch. */
    int dpi() const;
    
    /**
     * This is a convenience method that can be used to create a QImage from the image data
     * returned by the imageReady(...) signal. Note: If the image data has 16 bits/color the
     * data is truncated to 8 bits/color since QImage does not support 16 bits/color.
     * @param error is the pointer to string that will contain description of the error
     * if it occured during conversion
     * @param warning is the pointer to string that will contain warning
     * if it occured during conversion
     * @return This function returns the provided image data as a QImage. */
    QImage toQImage(QString* error = 0, QString* warning = 0) const;
    
private:
    QSharedDataPointer<SaneImageDataPrivate> d;
};

}

#endif
