/* ============================================================
*
* This file is part of the KDE project
*
* Date        : 2009-11-13
* Description : Sane interface for KDE
*
* Copyright (C) 2009 by Kare Sars <kare dot sars at iki dot fi>
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) version 3, or any
* later version accepted by the membership of KDE e.V. (or its
* successor approved by the membership of KDE e.V.), which shall
* act as a proxy defined in Section 6 of version 3 of the license.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
* ============================================================ */

// Local includes.
#include "sane_preview_thread.h"
#include "sane_preview_thread.moc"

#include <QDebug>

namespace LibQSane
{
    SanePreviewThread::SanePreviewThread(SANE_Handle handle, QImage *img):
    QThread(),
    status(SANE_STATUS_GOOD),
    m_frameSize(0),
    m_frameRead(0),
    m_dataSize(0),
    m_frame_t_count(0),
    m_saneHandle(handle),
    m_invertColors(false),
    m_readStatus(READ_READY),
    m_scanProgress(0),
    m_saneStartDone(false),
    m_imageUpdater(new PreviewImageUpdater(img))
    {
    }
    
    void SanePreviewThread::setPreviewInverted(bool inverted)
    {
        m_invertColors = inverted;
    }
    
    void SanePreviewThread::cancelScan()
    {
        m_readStatus = READ_CANCEL;
    }
    
    void SanePreviewThread::run()
    {
        m_dataSize = 0;
        m_readStatus = READ_ON_GOING;
        m_saneStartDone = false;

        // Start the scanning with sane_start
        status = sane_start(m_saneHandle);
        
        if (status != SANE_STATUS_GOOD) {
            qDebug() << "sane_start=" << sane_strstatus(status);
            sane_cancel(m_saneHandle);
            m_readStatus = READ_ERROR;
            return;
        }
        
        // Read image parameters
        status = sane_get_parameters(m_saneHandle, &m_params);
        if (status != SANE_STATUS_GOOD) {
            qDebug() << "sane_get_parameters=" << sane_strstatus(status);
            sane_cancel(m_saneHandle);
            m_readStatus = READ_ERROR;
            return;
        }

        qDebug() << Q_FUNC_INFO << "lines =" << m_params.lines;

        // calculate data size
        m_frameSize  = m_params.lines * m_params.bytes_per_line;
        if ((m_params.format == SANE_FRAME_RED) ||
            (m_params.format == SANE_FRAME_GREEN) ||
            (m_params.format == SANE_FRAME_BLUE))
        {
            // this is unfortunately calculated again for every frame....
            m_dataSize = m_frameSize*3;
        }
        else {
            m_dataSize = m_frameSize;
        }
        
        m_frameRead   = 0;
        m_frame_t_count = 0;

        m_imageUpdater->start(m_params);
        
        // set the m_saneStartDone here so the new QImage gets allocated before updating the preview.
        m_saneStartDone = true;
        
        while (m_readStatus == READ_ON_GOING) {
            readData();
        }
    }
    
    int SanePreviewThread::scanProgress()
    {
        // handscanners have negative data size
        if (m_dataSize <= 0) return 0;
        
        qint64 bytesRead;
        
        if (m_frameSize < m_dataSize) {
            bytesRead = m_frameRead + (m_frameSize * m_frame_t_count);
        }
        else {
            bytesRead = m_frameRead;
        }
        
        return (int)((bytesRead * 100)/m_dataSize);
    }
    
    void SanePreviewThread::readData()
    {
        SANE_Int readBytes;
        status = sane_read(m_saneHandle, m_readData, PREVIEW_READ_CHUNK_SIZE, &readBytes);
        
        switch (status) 
        {
            case SANE_STATUS_GOOD:
                // continue to parsing the data
                break;
                
            case SANE_STATUS_EOF:
                // handscanners have negative frame size (and not only handscanners) 
                if (m_frameRead < m_frameSize) {
                    qDebug() << "frameRead =" << m_frameRead  << ", frameSize =" << m_frameSize;
                    m_readStatus = READ_ERROR;
                    return;
                }
                if (m_params.last_frame == SANE_TRUE) {
                    // this is where it all ends well :)
                    m_readStatus = READ_READY;
                    return;
                }
                else {
                    // start reading next frame
                    SANE_Status status = sane_start(m_saneHandle);
                    if (status != SANE_STATUS_GOOD) {
                        qDebug() << "sane_start =" << sane_strstatus(status);
                        m_readStatus = READ_ERROR;
                        return;
                    }
                    status = sane_get_parameters(m_saneHandle, &m_params);
                    if (status != SANE_STATUS_GOOD) {
                        qDebug() << "sane_get_parameters =" << sane_strstatus(status);
                        m_readStatus = READ_ERROR;
                        sane_cancel(m_saneHandle);
                        return;
                    }
                    //qDebug() << "New Frame";
                    m_frameRead = 0;
                    m_frame_t_count++;
                    
                    m_imageUpdater->beginFrame(m_params);
                    
                    break;
                }
            default:
                qDebug() << "sane_read=" << status << "=" << sane_strstatus(status);
                m_readStatus = READ_ERROR;
                sane_cancel(m_saneHandle);
                return;
        }
        
        if (m_invertColors)
            invertReadData(readBytes);
        
        if (!m_imageUpdater->update(m_readData, readBytes))
            m_readStatus = READ_ERROR;
        else
        {
            m_frameRead += readBytes;
        }
    }
    
    void SanePreviewThread::invertReadData(int read_bytes)
    {
        if (m_params.depth >= 8) {
            for(int i=0; i<read_bytes; i++) {
                m_readData[i] = 255 - m_readData[i];
            }
        }
        else if (m_params.depth == 1) {
            for(int i=0; i<read_bytes; i++) {
                m_readData[i] = ~m_readData[i];
            }
        }
    }

    bool SanePreviewThread::saneStartDone()
    {
        return m_saneStartDone;
    }

}  // NameSpace LibQSane
