/* ============================================================
*
* This file is part of the KDE project
*
* Date        : 2008-11-15
* Description : Preview image viewer that can handle a selection.
*
* Copyright (C) 2008 by Kare Sars <kare dot sars at iki dot fi>
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) version 3, or any
* later version accepted by the membership of KDE e.V. (or its
* successor approved by the membership of KDE e.V.), which shall
* act as a proxy defined in Section 6 of version 3 of the license.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
* ============================================================ */

#ifndef SANE_VIEWER_H
#define SANE_VIEWER_H

#include "libqsane_export.h"

#include <QGraphicsView>
#include <QWheelEvent>

namespace LibQSane
{

class SaneViewerPrivate;

class LIBQSANE_EXPORT SaneViewer : public QGraphicsView
{
    Q_OBJECT
    public:
        explicit SaneViewer(QImage *img, QWidget *parent = 0);
        ~SaneViewer();

        void setQImage(QImage *img);
        void updateImage();
        /** Find selections in the picture
        * \param area this parameter determine the area of the reduced sized image. */
        void findSelections(float area = 10000.0);
        
        bool areSelectionsAllowed() const;
        bool areMultipleSelectionsAllowed() const;
        
        virtual QSize sizeHint() const;
        
    public Q_SLOTS:
        void setSelectionsAllowed(bool enabled = true);
        void setMultipleSelectionsAllowed(bool enabled = true);
        
        void setTLX(float ratio);
        void setTLY(float ratio);
        void setBRX(float ratio);
        void setBRY(float ratio);
        
        /** This function is used to set a selection without the user setting it.
        * \note all parameters must be in the range 0.0 -> 1.0.
        * \param tl_x is the x coordinate of the top left corner 0=0 1=image with.
        * \param tl_y is the y coordinate of the top left corner 0=0 1=image height.
        * \param br_x is the x coordinate of the bottom right corner 0=0 1=image with.
        * \param br_y is the y coordinate of the bottom right corner 0=0 1=image height. */
        void setSelection(float tl_x, float tl_y, float br_x, float br_y);
        void clearActiveSelection();
        void clearSavedSelections();
        void clearSelections();
        
        /** This function is used to darken everything except what is inside the given area.
        * \note all parameters must be in the range 0.0 -> 1.0.
        * \param tl_x is the x coordinate of the top left corner 0=0 1=image with.
        * \param tl_y is the y coordinate of the top left corner 0=0 1=image height.
        * \param br_x is the x coordinate of the bottom right corner 0=0 1=image with.
        * \param br_y is the y coordinate of the bottom right corner 0=0 1=image height. */
        void setHighlightArea(float tl_x, float tl_y, float br_x, float br_y);

        /** This function sets the percentage of the highlighted area that is visible. The rest is hidden.
        * \param percentage is the percentage of the highlighted area that is shown.
        * \param hideColor is the color to use to hide the highlighted area of the image.*/
        void setHighlightShown(int percentage, QColor hideColor = Qt::white);

        /** This function removes the highlight area. */
        void clearHighlight();

        void zoomIn();
        void zoomOut();
        void zoomSel();
        void zoom2Fit();
        void zoomOrig();

        int selListSize();
        /* This function returns the active visible selection in index 0 and after that the "saved" ones */
        bool selectionAt(int index, float &tl_x, float &tl_y, float &br_x, float &br_y);
        bool selectionRectAt(int index, QRect& rect);
        
    Q_SIGNALS:
        void newSelection(float tl_x, float tl_y, float br_x, float br_y);
        
    protected:
        void wheelEvent(QWheelEvent *e);
        void mousePressEvent(QMouseEvent *e);
        void mouseReleaseEvent(QMouseEvent *e);
        void mouseMoveEvent(QMouseEvent *e);
        void drawBackground(QPainter *painter, const QRectF &rect);
        
    private:
        QScopedPointer<SaneViewerPrivate> d_ptr;
        Q_DECLARE_PRIVATE(SaneViewer);
};

}  // NameSpace LibQSane


#endif


