/* ============================================================
 *
 * This file is part of the KDE project
 *
 * Date        : 2009-01-24
 * Description : Sane interface for KDE
 *
 * Copyright (C) 2007-2010 by Kare Sars <kare dot sars at iki dot fi>
 * Copyright (C) 2009 by Matthias Nagl <matthias at nagl dot info>
 * Copyright (C) 2009 by Grzegorz Kurtyka <grzegorz dot kurtyka at gmail dot com>
 * Copyright (C) 2007-2008 by Gilles Caulier <caulier dot gilles at gmail dot com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) version 3, or any
 * later version accepted by the membership of KDE e.V. (or its
 * successor approved by the membership of KDE e.V.), which shall
 * act as a proxy defined in Section 6 of version 3 of the license.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * ============================================================ */

#include "sane_widget.h"
#include "sane_widget.moc"

#include <unistd.h>

// Qt includes
#include <QApplication>
#include <QVarLengthArray>
#include <QLabel>
#include <QSplitter>
#include <QMutex>
#include <QPointer>
#include <QMessageBox>
#include <QHBoxLayout>

#include "config.h"

// Local includes.
#include "sane_widget_private.h"
#include "sane_option.h"
#include "sane_opt_button.h"
#include "sane_opt_checkbox.h"
#include "sane_opt_combo.h"
#include "sane_opt_entry.h"
#include "sane_opt_fslider.h"
#include "sane_opt_gamma.h"
#include "sane_opt_slider.h"
#include "labeled_separator.h"
#include "labeled_gamma.h"
#include "sane_device_dialog.h"
#include "sane_password_dialog.h"
#include "sane_options_widget.h"
#include "sane_auth.h"

namespace LibQSane
{
static int     s_objectCount = 0;
static QMutex  s_objectMutex;

SaneWidget::SaneWidget(QWidget* parent)
    : QWidget(parent), d_ptr(new SaneWidgetPrivate)
{
    Q_D(SaneWidget);
    d->q_ptr = this;

    SANE_Int    version;
    SANE_Status status;

    //qDebug() <<  "The language is:" << KGlobal::locale()->language();
    //qDebug() <<  "Languagelist" << KGlobal::locale()->languageList();
    //KGlobal::locale()->insertCatalog("libksane");
    //KGlobal::locale()->insertCatalog("sane-backends");

    s_objectMutex.lock();
    s_objectCount++;
    
    if (s_objectCount == 1) {
        status = sane_init(&version, &SaneAuth::authorization);
        if (status != SANE_STATUS_GOOD) {
            qDebug() << "libqsane: sane_init() failed("
                    << sane_strstatus(status) << ")";
        }
        else {
            //qDebug() << "Sane Version = "
            //         << SANE_VERSION_MAJOR(version) << "."
            //         << SANE_VERSION_MINORparent(version) << "."
            //         << SANE_VERSION_BUILD(version);
        }
    }
    s_objectMutex.unlock();

    d->m_updProgressTmr.setSingleShot(false);
    d->m_updProgressTmr.setInterval(300);
    connect(&d->m_updProgressTmr, SIGNAL(timeout()), d, SLOT(updateProgress()));
    
    // Forward signals from the private class
    connect(d, SIGNAL(scanProgress(int)), this, SIGNAL(scanProgress(int)));
    connect(d, SIGNAL(scanDone(int, QString)), this, SIGNAL(scanDone(int, QString)));
    
    // Create the static UI
    // create the preview
    d->m_previewViewer = new SaneViewer(&(d->m_previewImg), this);
    connect(d->m_previewViewer, SIGNAL(newSelection(float, float, float, float)),
            d, SLOT(handleSelection(float, float, float, float)));

    
    d->m_warmingUp = new QLabel;
    d->m_warmingUp->setText(tr("Waiting for the scan to start."));
    d->m_warmingUp->setAlignment(Qt::AlignCenter);
    d->m_warmingUp->setAutoFillBackground(true);
    d->m_warmingUp->setBackgroundRole(QPalette::Highlight);
    //d->m_warmingUp->setForegroundRole(QPalette::HighlightedText);
    d->m_warmingUp->hide();

    d->m_progressBar = new QProgressBar;
    d->m_progressBar->setMaximum(100);

    d->m_cancelBtn   = new QPushButton;
    d->m_cancelBtn->setIcon(QIcon::fromTheme("process-stop", QIcon(ICONS_INSTALL_DIR"/process-stop.png")));
    d->m_cancelBtn->setToolTip(tr("Cancel current scan operation"));
    connect(d->m_cancelBtn, SIGNAL(clicked()), this, SLOT(scanCancel()));
    
    d->m_activityFrame = new QWidget;
    QHBoxLayout *progress_lay = new QHBoxLayout(d->m_activityFrame);
    progress_lay->setContentsMargins(0,0,0,0);
    progress_lay->addWidget(d->m_progressBar, 100);
    progress_lay->addWidget(d->m_cancelBtn, 0);
    d->m_activityFrame->hide();
    
    d->m_zInBtn  = new QToolButton(this);
    d->m_zInBtn->setIcon(QIcon::fromTheme("zoom-in", QIcon(ICONS_INSTALL_DIR"/zoom-in.png")));
    d->m_zInBtn->setToolTip(tr("Zoom In"));
    connect(d->m_zInBtn, SIGNAL(clicked()), d->m_previewViewer, SLOT(zoomIn()));
    
    d->m_zOutBtn = new QToolButton(this);
    d->m_zOutBtn->setIcon(QIcon::fromTheme("zoom-out", QIcon(ICONS_INSTALL_DIR"/zoom-out.png")));
    d->m_zOutBtn->setToolTip(tr("Zoom Out"));
    connect(d->m_zOutBtn, SIGNAL(clicked()), d->m_previewViewer, SLOT(zoomOut()));
    
    d->m_zSelBtn = new QToolButton(this);
    d->m_zSelBtn->setIcon(QIcon::fromTheme("document-preview", QIcon(ICONS_INSTALL_DIR"/document-preview.png")));
    d->m_zSelBtn->setToolTip(tr("Zoom to Selection"));
    connect(d->m_zSelBtn, SIGNAL(clicked()), d->m_previewViewer, SLOT(zoomSel()));
    
    d->m_zFitBtn = new QToolButton(this);
    d->m_zFitBtn->setIcon(QIcon::fromTheme("zoom-fit-best", QIcon(ICONS_INSTALL_DIR"/zoom-fit-best.png")));
    d->m_zFitBtn->setToolTip(tr("Zoom to Fit"));
    connect(d->m_zFitBtn, SIGNAL(clicked()), d->m_previewViewer, SLOT(zoom2Fit()));
    
    /*d->m_zOrigBtn = new QPushButton(this);
    d->m_zOrigBtn->setIcon(QIcon::fromTheme("zoom-original", QIcon(ICONS_INSTALL_DIR"/zoom-original.png")));
    d->m_zOrigBtn->setToolTip(tr("Original size"));
    connect(d->m_zOrigBtn, SIGNAL(clicked()), d->m_previewViewer, SLOT(zoomOrig()));*/
    
    d->m_prevBtn = new QPushButton(this);
    d->m_prevBtn->setToolTip(tr("Scan Preview Image"));
    //: Preview button text
    d->m_prevBtn->setText(tr("Preview"));
    connect(d->m_prevBtn,   SIGNAL(clicked()), d, SLOT(startPreviewScan()));
    
    d->m_scanBtn = new QPushButton(this);
    d->m_scanBtn->setToolTip(tr("Scan Final Image"));
    //: Final scan button text
    d->m_scanBtn->setText(tr("Scan"));
    d->m_scanBtn->setFocus(Qt::OtherFocusReason);
    connect(d->m_scanBtn,   SIGNAL(clicked()), d, SLOT(startFinalScan()));
    
    d->m_btnFrame = new QWidget;
    QHBoxLayout *btn_lay = new QHBoxLayout(d->m_btnFrame);
    btn_lay->setContentsMargins(0,0,0,0);
    btn_lay->addWidget(d->m_zInBtn);
    btn_lay->addWidget(d->m_zOutBtn);
    //btn_lay->addWidget(d->m_zOrigBtn);
    btn_lay->addWidget(d->m_zFitBtn);
    btn_lay->addWidget(d->m_zSelBtn);
    btn_lay->addStretch(100);
    btn_lay->addWidget(d->m_prevBtn);
    btn_lay->addWidget(d->m_scanBtn);
    
    // calculate the height of the waiting/scanning/buttons frames to avoid jumpiness.
    int minHeight = d->m_btnFrame->sizeHint().height();
    if (d->m_activityFrame->sizeHint().height() > minHeight) minHeight = d->m_activityFrame->sizeHint().height();
    if (d->m_warmingUp->sizeHint().height() > minHeight)     minHeight = d->m_warmingUp->sizeHint().height();
    d->m_btnFrame->setMinimumHeight(minHeight);
    d->m_activityFrame->setMinimumHeight(minHeight);
    d->m_warmingUp->setMinimumHeight(minHeight);
    
    d->m_previewFrame = new QWidget;
    QVBoxLayout *preview_layout = new QVBoxLayout(d->m_previewFrame);
    preview_layout->setContentsMargins(0,0,0,0);
    preview_layout->addWidget(d->m_previewViewer, 100);
    preview_layout->addWidget(d->m_warmingUp, 0);
    preview_layout->addWidget(d->m_activityFrame, 0);
    preview_layout->addWidget(d->m_btnFrame, 0);
    
    // Create Options Widget
    d->m_optsWidget = new SaneOptionsWidget;
    connect(d->m_optsWidget, SIGNAL(optionsReloaded()), d, SLOT(optReload()));
    connect(d->m_optsWidget, SIGNAL(invertColorsToggled(bool)), d, SLOT(invertPreview()));
    
    d->m_splitter = new QSplitter(this);
    d->m_splitter->addWidget(d->m_optsWidget);
    d->m_splitter->setStretchFactor(0,0);
    d->m_splitter->addWidget(d->m_previewFrame);
    d->m_splitter->setStretchFactor(1,100);
    
    d->m_optionsCollapser = new SplitterCollapser(d->m_splitter, d->m_optsWidget);
    
    QHBoxLayout *base_layout = new QHBoxLayout(this);
    base_layout->addWidget(d->m_splitter);
    base_layout->setContentsMargins(0,0,0,0);

    // disable the interface in case no device is opened.
    d->m_optsWidget->setDisabled(true);
    d->m_previewViewer->setDisabled(true);
    d->m_btnFrame->setDisabled(true);
    
}

SaneWidget::~SaneWidget()
{
    qDebug() << Q_FUNC_INFO << "{";
    Q_D(SaneWidget);
    
    while (!closeDevice()) usleep(1000);
    // wait for any thread to exit

    s_objectMutex.lock();
    s_objectCount--;
    if (s_objectCount <= 0) {
        // only delete the authorization singleton and call sane_exit
        delete d->m_auth;
        sane_exit();
    }
    s_objectMutex.unlock();

    qDebug() << Q_FUNC_INFO << "}";
}

QString SaneWidget::device() const {return d_func()->m_deviceInfo.name;}
QString SaneWidget::vendor() const {return d_func()->m_deviceInfo.vendor;}
QString SaneWidget::make() const {return d_func()->m_deviceInfo.vendor;}
QString SaneWidget::model() const {return d_func()->m_deviceInfo.model;}

QList<SaneDeviceInfo> SaneWidget::getDevices()
{
    QList<SaneDeviceInfo> devices;

    SaneDeviceInfo deviceInfo;

    // get the device list to get the vendor and model info
    SANE_Device const **dev_list;
    SANE_Status status = sane_get_devices(&dev_list, SANE_FALSE);
    if (status != SANE_STATUS_GOOD)
        return devices;

    int i = 0;
    while (dev_list[i] != 0)
    {
        deviceInfo.name   = QString::fromLocal8Bit(dev_list[i]->name);
        deviceInfo.vendor = QString::fromLocal8Bit(dev_list[i]->vendor);
        deviceInfo.model  = QString::fromLocal8Bit(dev_list[i]->model);
        deviceInfo.type   = QString::fromLocal8Bit(dev_list[i]->type);

        devices.append(deviceInfo);

        i++;
    }

    return devices;
}

SaneDeviceInfo SaneWidget::selectDevice(QWidget* parent)
{
  SaneDeviceInfo selected_device;
  QScopedPointer<SaneDeviceDialog> sel(new SaneDeviceDialog(parent));

  // sel.setDefault(prev_backend); // set default scanner - perhaps application using libqsane should remember that
  if(sel->exec() == QDialog::Accepted) {
      selected_device = sel->getSelectedDevice();
  }
  
  return selected_device;
}

bool SaneWidget::openDevice(const SaneDeviceInfo &deviceInfo)
{
    Q_D(SaneWidget);
    
    QString deviceName = deviceInfo.name;
    
    if (deviceName.isEmpty()) {
        return false;
    }
    
    if (deviceName == "test")
    {
        d->m_deviceInfo.name   = "test";
        d->m_deviceInfo.vendor = "Test";
        d->m_deviceInfo.model  = "Scanner";
        d->m_deviceInfo.type   = "test scanner";
    }
    else
    {
        d->m_deviceInfo = deviceInfo;
    }

    // Try to open the device
    SANE_Status status = sane_open(deviceName.toLatin1(), &d->m_saneHandle);

    bool password_dialog_ok = true;

    SanePasswordDialog* passwordDialog;
    // prepare wallet for authentication and create password dialog
    if (status == SANE_STATUS_ACCESS_DENIED) {
        /*sane_wallet = KWallet::Wallet::openWallet(KWallet::Wallet::LocalWallet(), winId() );

        if(sane_wallet) {
            dlg = new KPasswordDialog(this, KPasswordDialog::ShowUsernameLine | KPasswordDialog::ShowKeepPassword);
            if(!sane_wallet->hasFolder(my_folder_name)) {
                sane_wallet->createFolder(my_folder_name);
            }
            sane_wallet->setFolder(my_folder_name);
            sane_wallet->readMap(deviceName.toLatin1(), wallet_entry);
            if(!wallet_entry.empty() || true) {
                dlg->setUsername( wallet_entry["username"] );
                dlg->setPassword( wallet_entry["password"] );
                dlg->setKeepPassword( true );
            }
        } else {*/
            passwordDialog = new SanePasswordDialog(this);// KPasswordDialog::ShowUsernameLine);
        //}
        passwordDialog->setPrompt(tr("Authentication required for resource: %1").arg( deviceName ) );

    }

    // sane_open failed due to insufficient authorization
    // retry opening device with user provided data assisted with kwallet records
    while (status == SANE_STATUS_ACCESS_DENIED) {

        password_dialog_ok = passwordDialog->exec();
        if (!password_dialog_ok) {
            delete passwordDialog;
            return false; //the user canceled
        }

        d->m_auth->setDeviceAuth(deviceName, passwordDialog->username(), passwordDialog->password());

        status = sane_open(deviceName.toLatin1(), &d->m_saneHandle);

        // store password in wallet on successful authentication
        /*if(dlg->keepPassword() && status != SANE_STATUS_ACCESS_DENIED) {
            QMap<QString, QString> entry;
            entry["username"] = dlg->username().toUtf8();
            entry["password"] = dlg->password().toUtf8();
            sane_wallet->writeMap(deviceName.toLatin1(), entry);
        }*/
    }

    if (status != SANE_STATUS_GOOD) {
        qDebug() << "sane_open(\"" << deviceName << "\", &handle) failed! status = " << sane_strstatus(status);
        d->m_auth->clearDeviceAuth(deviceName);
        return false;
    }

    // Read the options (start with option 0 the number of parameters)
    const SANE_Option_Descriptor* optDesc = sane_get_option_descriptor(d->m_saneHandle, 0);
    if (optDesc == 0) {
        d->m_auth->clearDeviceAuth(deviceName);
        return false;
    }
    QVarLengthArray<char> data(optDesc->size);
    SANE_Int res;
    status = sane_control_option(d->m_saneHandle, 0, SANE_ACTION_GET_VALUE, data.data(), &res);
    if (status != SANE_STATUS_GOOD) {
        d->m_auth->clearDeviceAuth(deviceName);
        return false;
    }
    
    SANE_Word numSaneOptions = *reinterpret_cast<SANE_Word*>(data.data());

    QList<SaneOption*> optList;
    
    // read the rest of the options
    for (int i=1; i<numSaneOptions; ++i)
    {
        const SANE_Option_Descriptor* optionDescriptor = sane_get_option_descriptor(d->m_saneHandle, i);
        qDebug() << Q_FUNC_INFO << "option name=" << optionDescriptor->name << ", type =" << SaneOption::optionType(optionDescriptor) << ", size =" << optionDescriptor->size;
        switch (SaneOption::optionType(optionDescriptor)) {
            case SaneOption::TYPE_DETECT_FAIL:
                optList.append(new SaneOption(d->m_saneHandle, i));
                break;
            case SaneOption::TYPE_CHECKBOX:
                optList.append(new SaneOptCheckBox(d->m_saneHandle, i));
                break;
            case SaneOption::TYPE_SLIDER:
                optList.append(new SaneOptSlider(d->m_saneHandle, i));
                break;
            case SaneOption::TYPE_F_SLIDER:
                optList.append(new SaneOptFSlider(d->m_saneHandle, i));
                break;
            case SaneOption::TYPE_COMBO:
                optList.append(new SaneOptCombo(d->m_saneHandle, i));
                break;
            case SaneOption::TYPE_ENTRY:
                optList.append(new SaneOptEntry(d->m_saneHandle, i));
                break;
            case SaneOption::TYPE_GAMMA:
                optList.append(new SaneOptGamma(d->m_saneHandle, i));
                break;
            case SaneOption::TYPE_BUTTON:
                optList.append(new SaneOptButton(d->m_saneHandle, i));
                break;
        }
    }
    
    // Create the preview thread
    d->m_previewThread = new SanePreviewThread(d->m_saneHandle, &d->m_previewImg);
    connect(d->m_previewThread, SIGNAL(finished()), d, SLOT(previewScanDone()));
    
    // Create the read thread
    d->m_scanThread = new SaneScanThread(d->m_saneHandle, &d->m_scanData, &d->m_previewImg);
    connect(d->m_scanThread, SIGNAL(finished()), d, SLOT(oneFinalScanDone()));
    
    // Create the options interface
    d->m_optsWidget->createOptInterface(optList, d->m_saneHandle);
    connect(d->m_optsWidget, SIGNAL(tlXChanged(float)), d, SLOT(setTLX(float)));
    connect(d->m_optsWidget, SIGNAL(tlYChanged(float)), d, SLOT(setTLY(float)));
    connect(d->m_optsWidget, SIGNAL(brXChanged(float)), d, SLOT(setBRX(float)));
    connect(d->m_optsWidget, SIGNAL(brYChanged(float)), d, SLOT(setBRY(float)));
    
    d->m_previewViewer->setSelectionsAllowed(d->m_optsWidget->canSetScanArea());
    d->m_previewViewer->setMultipleSelectionsAllowed(!d->m_optsWidget->isADFSource());
    connect(d->m_optsWidget, SIGNAL(scanSourceChanged()), d, SLOT(scanSourceChanged()));
    d->m_zSelBtn->setVisible(d->m_optsWidget->canSetScanArea());
    
    // try to set SaneWidget default values
    d->m_optsWidget->setDefaultValues();
    
    // Enable the interface
    d->m_optsWidget->setDisabled(false);
    d->m_previewViewer->setDisabled(false);
    d->m_btnFrame->setDisabled(false);

    // estimate the preview size and create an empty image
    // this is done so that you can select scan area without
    // having to scan a preview.
    d->updatePreviewSize();
    // it's a dirty hack, view doesn't have time to become visible
    //QTimer::singleShot(0, d->m_previewViewer, SLOT(zoom2Fit()));
    return true;
}

bool SaneWidget::openDevice(const QString &deviceName)
{
    // don't bother trying to open if the device string is empty
    if (deviceName.isEmpty()) {
        return false;
    }

    if (deviceName == "test")
    {
        SaneDeviceInfo deviceInfo;
        deviceInfo.name = "test";
        return openDevice(deviceInfo);
    }
    else
    {
        foreach (SaneDeviceInfo deviceInfo, getDevices())
        {
            if (deviceInfo.name == deviceName)
                return openDevice(deviceInfo);
        }
    }

    return false;
}

bool SaneWidget::closeDevice()
{
    Q_D(SaneWidget);
    if (!d->m_saneHandle) {
        return true;
    }
    
    if (d->m_scanThread->isRunning()) {
        d->m_scanThread->cancelScan();
        d->m_closeDevicePending = true;
        return false;
    }
    
    if (d->m_previewThread->isRunning()) {
        d->m_previewThread->cancelScan();
        d->m_closeDevicePending = true;
        return false;
    }
    
    d->m_auth->clearDeviceAuth(d->m_deviceInfo.name);
    
    // else 
    sane_close(d->m_saneHandle);
    d->m_saneHandle = 0;
    d->clearDeviceOptions();

    // disable the interface until a new device is opened.
    d->m_optsWidget->setDisabled(true);
    d->m_previewViewer->setDisabled(true);
    d->m_btnFrame->setDisabled(true);
    
    return true;
}

void SaneWidget::scanFinal()
{
    Q_D(SaneWidget);
    if (d->m_btnFrame->isEnabled()) {
        d->startFinalScan();
    }
    else {
        // if the button frame is disabled, there is no open device to scan from
        emit scanDone(SaneWidget::ErrorGeneral, "");
    }
}

void SaneWidget::scanCancel()
{
    Q_D(SaneWidget);
    if (d->m_scanThread->isRunning()) {
        d->m_scanThread->cancelScan();
    }
    else
        d->m_cancelNextScan = true; // supposed that this method was called immediately after imageReady signal
    
    if (d->m_previewThread->isRunning()) {
        d->m_previewThread->cancelScan();
    }
}

void SaneWidget::setPreviewResolution(float dpi)
{
    Q_D(SaneWidget);
    d->m_previewDPI = dpi;
}

QMap<QString, QByteArray> SaneWidget::getOptVals()
{
    Q_D(SaneWidget);
    return d->m_optsWidget->getOptVals();
}

bool SaneWidget::getOptVal(const QString &optname, QByteArray &value)
{
    Q_D(SaneWidget);
    return d->m_optsWidget->getOptVal(optname, value);
}

int SaneWidget::setOptVals(const QMap <QString, QByteArray> &opts)
{
    Q_D(SaneWidget);
    return d->m_optsWidget->setOptVals(opts);
}

bool SaneWidget::setOptVal(const QString &option, const QByteArray &value)
{
    Q_D(SaneWidget);
    return d->m_optsWidget->setOptVal(option, value);
}

void SaneWidget::setScanButtonText(const QString &scanLabel)
{
    Q_D(SaneWidget);
    if (d->m_scanBtn == 0) {
        qCritical() << "setScanButtonText was called before SaneWidget was initialized";
        return;
    }
    d->m_scanBtn->setText(scanLabel);
}

void SaneWidget::setScanButtonIcon(const QIcon &scanIcon)
{
    Q_D(SaneWidget);
    if (d->m_scanBtn == 0) {
        qCritical() << "setScanButtonIcon was called before SaneWidget was initialized";
        return;
    }
    d->m_scanBtn->setIcon(scanIcon);
}

void SaneWidget::setPreviewButtonText(const QString &previewLabel)
{
    Q_D(SaneWidget);
    if (d->m_scanBtn == 0) {
        qCritical() << "setPreviewButtonText was called before SaneWidget was initialized";
        return;
    }
    d->m_prevBtn->setText(previewLabel);
}

void SaneWidget::setPreviewButtonIcon(const QIcon &previewIcon)
{
    Q_D(SaneWidget);
    if (d->m_scanBtn == 0) {
        qCritical() << "setPreviewButtonIcon was called before SaneWidget was initialized";
        return;
    }
    d->m_prevBtn->setIcon(previewIcon);
}

void SaneWidget::enableAutoSelect(bool enable)
{
    Q_D(SaneWidget);
    d->m_autoSelect = enable;
}

float SaneWidget::scanAreaWidth()
{
    Q_D(SaneWidget);
    return d->m_optsWidget->scanAreaWidth();
}

float SaneWidget::scanAreaHeight()
{
    Q_D(SaneWidget);
    return d->m_optsWidget->scanAreaHeight();
}

// topLeft and bottomRight are in mm
void SaneWidget::setSelection(QPointF topLeft, QPointF bottomRight)
{
    Q_D(SaneWidget);
    
    if (!d->m_optsWidget->canSetScanArea())
        return;
    
    if (!d->m_optsWidget->setScanArea(topLeft, bottomRight))
    {
        d->m_previewViewer->clearActiveSelection();
        d->m_optsWidget->setFullScanArea();
    }
    else
    {
        float xmax = d->m_optsWidget->scanAreaWidth();
        float ymax = d->m_optsWidget->scanAreaHeight();
        if (xmax != 0.0 && ymax != 0.0)
            d->m_previewViewer->setSelection(topLeft.x() / xmax, topLeft.y() / ymax, bottomRight.x() / xmax, bottomRight.y() / ymax);
        else
            d->m_previewViewer->clearActiveSelection();
    }
}

void SaneWidget::setOptionsCollapsed(bool collapse)
{
    Q_D(SaneWidget);
    if (collapse) {
        QTimer::singleShot(0, d->m_optionsCollapser, SLOT(slotCollapse()));
    }
    else {
        QTimer::singleShot(0, d->m_optionsCollapser, SLOT(slotRestore()));
    }
}

void SaneWidget::setScanButtonHidden(bool hidden)
{
    Q_D(SaneWidget);
    d->m_scanBtn->setHidden(hidden);
}

void SaneWidget::setMultipleSelectionsAllowed(bool enable)
{
    Q_D(SaneWidget);
    
    d->m_previewViewer->setMultipleSelectionsAllowed(enable);
}

}  // NameSpace LibQSane
