/* ============================================================
 *
 * This file is part of the KDE project
 *
 * Date        : 2007-09-13
 * Description : Sane interface for KDE
 *
 * Copyright (C) 2007-2008 by Kare Sars <kare dot sars at iki dot fi>
 * Copyright (C) 2007-2008 by Gilles Caulier <caulier dot gilles at gmail dot com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) version 3, or any
 * later version accepted by the membership of KDE e.V. (or its
 * successor approved by the membership of KDE e.V.), which shall
 * act as a proxy defined in Section 6 of version 3 of the license.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * ============================================================ */

// Local includes.
#include "sane_widget_private.h"
#include "sane_widget_private.moc"

// Qt includes.
#include <QImage>
#include <QScrollArea>
#include <QScrollBar>
#include <QList>
#include <QLabel>
#include <QMessageBox>
#include <QPushButton>
#include <QVBoxLayout>

#include "sane_options_widget.h"
#include "sane_option.h"
#include "sane_auth.h"

#define SCALED_PREVIEW_MAX_SIDE 400

static const int ActiveSelection = 100000;

namespace LibQSane
{

SaneWidgetPrivate::SaneWidgetPrivate()
{
    // Device independent UI variables
    m_optsWidget = 0;
    m_zInBtn        = 0;
    m_zOutBtn       = 0;
    m_zSelBtn       = 0;
    m_zFitBtn       = 0;
    //m_zOrigBtn      = 0;
    m_prevBtn       = 0;
    m_scanBtn       = 0;
    m_cancelBtn     = 0;
    m_previewViewer = 0;
    m_autoSelect    = true;
    m_selIndex      = ActiveSelection;
    m_warmingUp     = 0;
    m_progressBar   = 0;

    // scanning variables
    m_isPreview     = false;

    m_saneHandle    = 0;
    m_previewThread = 0;
    m_scanThread    = 0;

    m_previewDPI    = 0;
    
    m_previewWidth  = 0;
    m_previewHeight = 0;
    
    clearDeviceOptions();
    
    m_auth = SaneAuth::getInstance();
}

void SaneWidgetPrivate::clearDeviceOptions()
{
    m_scanOngoing   = false;
    m_closeDevicePending = false;
    m_cancelNextScan = false;
    
    if (m_optsWidget)
        m_optsWidget->clearDeviceOptions();

    // remove the remaining layouts/widgets and read thread

    delete m_previewThread;
    m_previewThread = 0;

    delete m_scanThread;
    m_scanThread = 0;
}

SaneImageData::ImageFormat SaneWidgetPrivate::getImgFormat(SANE_Parameters &params)
{
    switch(params.format)
    {
        case SANE_FRAME_GRAY:
            switch(params.depth)
            {
                case 1:
                    return SaneImageData::FormatBlackWhite;
                case 8:
                    return SaneImageData::FormatGrayScale8;
                case 16:
                    return SaneImageData::FormatGrayScale16;
                default:
                    return SaneImageData::FormatNone;
            }
        case SANE_FRAME_RGB:
        case SANE_FRAME_RED:
        case SANE_FRAME_GREEN:
        case SANE_FRAME_BLUE:
            switch (params.depth)
            {
                case 8:
                    return SaneImageData::FormatRGB_8_C;
                case 16:
                    return SaneImageData::FormatRGB_16_C;
                default:
                    return SaneImageData::FormatNone;
            }
}
    return SaneImageData::FormatNone;
}

int SaneWidgetPrivate::getBytesPerLines(SANE_Parameters &params)
{
    switch (getImgFormat(params))
    {
        case SaneImageData::FormatBlackWhite:
        case SaneImageData::FormatGrayScale8:
        case SaneImageData::FormatGrayScale16:
            return params.bytes_per_line;

        case SaneImageData::FormatRGB_8_C:
            return params.pixels_per_line*3;

        case SaneImageData::FormatRGB_16_C:
            return params.pixels_per_line*6;

        case SaneImageData::FormatNone:
        case SaneImageData::FormatBMP: // to remove warning (BMP is omly valid in the twain wrapper)
            return 0;
    }
    return 0;
}

void SaneWidgetPrivate::clearPreviewImage()
{
    m_previewImg.fill(0xFFFFFFFF);
}

void SaneWidgetPrivate::optReload()
{
    qDebug() << Q_FUNC_INFO;
    // estimate the preview size and create an empty image
    // this is done so that you can select scan area without
    // having to scan a preview.
    updatePreviewSize();
    
    //m_previewViewer->zoom2Fit();
}

void SaneWidgetPrivate::handleSelection(float tl_x, float tl_y, float br_x, float br_y)
{
    if (!m_optsWidget->canSetScanArea())
    {
        // clear the selection since we can not set one
        m_previewViewer->clearActiveSelection();
        return;
    }
    
    if ((m_previewImg.width()==0) || (m_previewImg.height()==0)) return;
    
    m_optsWidget->setScanAreaByRatios(tl_x, tl_y, br_x, br_y);
}

void SaneWidgetPrivate::setTLX(float ftlx)
{
    // ignore this during an active scan
    if (m_previewThread->isRunning()) return;
    if (m_scanThread->isRunning()) return;
    if (m_scanOngoing) return;
    
    m_previewViewer->setTLX(ftlx);
}

void SaneWidgetPrivate::setTLY(float ftly)
{
    // ignore this during an active scan
    if (m_previewThread->isRunning()) return;
    if (m_scanThread->isRunning()) return;
    if (m_scanOngoing) return;
    
    m_previewViewer->setTLY(ftly);
}

void SaneWidgetPrivate::setBRX(float fbrx)
{
    // ignore this during an active scan
    if (m_previewThread->isRunning()) return;
    if (m_scanThread->isRunning()) return;
    if (m_scanOngoing) return;
    
    m_previewViewer->setBRX(fbrx);
}

void SaneWidgetPrivate::setBRY(float fbry)
{
    // ignore this during an active scan
    if (m_previewThread->isRunning()) return;
    if (m_scanThread->isRunning()) return;
    if (m_scanOngoing) return;
    
    m_previewViewer->setBRY(fbry);
}

void SaneWidgetPrivate::updatePreviewSize()
{   
    qDebug() << Q_FUNC_INFO;
    // check if an update is necessary
    float max_x = m_optsWidget->scanAreaWidth();
    float max_y = m_optsWidget->scanAreaHeight();
    
    if ((max_x == m_previewWidth) && (max_y == m_previewHeight))
    {
        //qDebug() << "no preview size change";
        return;
    }
    
    // The preview size has changed 
    m_previewWidth  = max_x;
    m_previewHeight = max_y;

    // set the scan area to the whole area
    m_previewViewer->clearSelections();
    m_optsWidget->setFullScanArea();
    
    // create a "scaled" image of the preview
    float ratio = max_x/max_y;
    int x,y;
    if (ratio < 1) {
        x=SCALED_PREVIEW_MAX_SIDE;
        y=(int)(SCALED_PREVIEW_MAX_SIDE/ratio);
    }
    else {
        y=SCALED_PREVIEW_MAX_SIDE;
        x=(int)(SCALED_PREVIEW_MAX_SIDE/ratio);
    }
    
    m_previewImg = QImage(x, y, QImage::Format_RGB32);
    clearPreviewImage();
    
    // set the new image
    m_previewViewer->setQImage(&m_previewImg);
}

void SaneWidgetPrivate::startPreviewScan()
{
    if (m_scanOngoing) return;
    m_scanOngoing = true;
        
    // store the current settings of parameters to be changed
    m_optsWidget->storePreviewOptions();
    
    // check if we can modify the selection
    if (m_optsWidget->canSetScanArea())
    {
        m_optsWidget->setFullScanArea();
    }
    else {
        // no use to try auto selections if you can not use them
        m_autoSelect = false;
    }
    
    if (!m_optsWidget->setScanResolution(m_previewDPI))
    {
        previewScanDone();
        return;
    }
    
    // set preview option to true if possible
    m_optsWidget->setPreviewMode();
    
    // execute valReload if there is a pending value reload
    m_optsWidget->flushValuesReloading();
    
    // clear the preview
    m_previewViewer->clearHighlight();
    m_previewViewer->clearSelections();
    clearPreviewImage();
    updatePreviewSize();
    
    setBusy(true);
    
    m_progressBar->setValue(0);
    m_isPreview = true;
    m_previewThread->setPreviewInverted(m_optsWidget->invertColors());
    m_previewThread->start();
    m_updProgressTmr.start();
}

void SaneWidgetPrivate::previewScanDone()
{
    // even if the scan is finished successfully we need to call sane_cancel()
    sane_cancel(m_saneHandle);
    
    if (m_closeDevicePending) {
        setBusy(false);
        sane_close(m_saneHandle);
        m_saneHandle = 0;
        clearDeviceOptions();
        emit scanDone(SaneWidget::NoError, "");
        return;
    }
    
    // restore the original settings of the changed parameters
    m_optsWidget->restorePreviewOptions();
    
    if (m_autoSelect) {
        m_previewViewer->findSelections();
    }
    
    setBusy(false);
    m_scanOngoing = false;
    m_cancelNextScan = false;
    m_updProgressTmr.stop();
    
    m_previewViewer->updateImage();
    
    emit scanDone(SaneWidget::NoError, "");
    
    return;
}


void SaneWidgetPrivate::startFinalScan()
{
    if (m_scanOngoing) return;
    m_scanOngoing = true;
    m_isPreview = false;
    
    float x1=0,y1=0, x2=0,y2=0;

    m_selIndex = 0;
    
    QRect scanRect = m_previewImg.rect();
    if (m_optsWidget->canSetScanArea())
    {
        // reead the selection from the viewer
        m_previewViewer->selectionAt(m_selIndex, x1,y1,x2,y2);
        qDebug() << Q_FUNC_INFO << "setting selection:" << x1 << y1 << x2 << y2;
        m_previewViewer->setHighlightArea(x1,y1,x2,y2);
 
        m_optsWidget->setScanAreaByRatios(x1, y1, x2, y2);
        
        m_previewViewer->selectionRectAt(m_selIndex, scanRect);
        
        m_selIndex++;
    }
    
    // execute a pending value reload
    m_optsWidget->flushValuesReloading();
    
    setBusy(true);
    m_updProgressTmr.start();
    m_scanThread->setImageInverted(m_optsWidget->invertColors());

    clearPreviewImage();
    
    float scanDPI = 0;
    m_optsWidget->getScanResolution(&scanDPI);
    
    m_scanThread->startScan(scanRect, scanDPI);
}

void SaneWidgetPrivate::oneFinalScanDone()
{
    Q_Q(SaneWidget);
    
    m_updProgressTmr.stop();
    
    if (m_closeDevicePending) {
        setBusy(false);
        sane_close(m_saneHandle);
        m_saneHandle = 0;
        clearDeviceOptions();
        return;
    }
    
    if (m_scanThread->frameStatus() == SaneScanThread::READ_READY) 
    {
        // scan finished OK
        SANE_Parameters params = m_scanThread->saneParameters();
        int lines = params.lines;
        if (lines == -1) {
            // this is probably a handscanner -> calculate the size from the read data
            int bytesPerLine = qMax(getBytesPerLines(params), 1); // ensure no div by 0
            lines = m_scanData.size() / bytesPerLine;
        }
        
        qDebug() << Q_FUNC_INFO << "image is ready";
        emit q->imageReady(SaneImageData(m_scanData,
                         params.pixels_per_line,
                         lines,
                         getBytesPerLines(params),
                         getImgFormat(params),
                         m_optsWidget->currentDPI()));

        if (m_cancelNextScan)
            sane_cancel(m_saneHandle);
        else
        {
            // now check if we should have automatic ADF batch scaning
            if (m_optsWidget->isADFSource())
            {
                // in batch mode only one area can be scanned per page
                //qDebug() << "source == \"Automatic Document Feeder\"";
                clearPreviewImage();
                m_updProgressTmr.start();
                m_scanThread->start();
                return;
            }

            // Check if we have a "wait for button" batch scanning
            if (m_optsWidget->waitForButton())
            {
                // in batch mode only one area can be scanned per page
                //qDebug() << "source == \"Automatic Document Feeder\"";
                clearPreviewImage();
                m_updProgressTmr.start();
                m_scanThread->start();
                return;
            }

            // not batch scan, call sane_cancel to be able to change parameters.
            sane_cancel(m_saneHandle);

            //qDebug() << "index=" << m_selIndex << "size=" << m_previewViewer->selListSize();
            // check if we have multiple selections.
            if (m_previewViewer->selListSize() > m_selIndex)
            {   
                if (m_optsWidget->canSetScanArea())
                {
                    float x1=0,y1=0, x2=0,y2=0;
                    
                    // read the selection from the viewer
                    m_previewViewer->selectionAt(m_selIndex, x1,y1,x2,y2);

                    // set the highlight
                    m_previewViewer->setHighlightArea(x1,y1,x2,y2);
                    
                    m_optsWidget->setScanAreaByRatios(x1, y1, x2, y2);
                    
                    QRect scanRect = m_previewImg.rect();
                    m_previewViewer->selectionRectAt(m_selIndex, scanRect);
                    
                    m_selIndex++;
                    
                    // execute a pending value reload
                    m_optsWidget->flushValuesReloading();
                    m_updProgressTmr.start();
                    
                    float scanDPI = 0;
                    m_optsWidget->getScanResolution(&scanDPI);
                    
                    m_scanThread->startScan(scanRect, scanDPI);
                    return;
                }
            }
        }
        
        emit scanDone(SaneWidget::NoError, "");
    }
    else {
        switch(m_scanThread->saneStatus()) 
        {
            case SANE_STATUS_GOOD:
            case SANE_STATUS_UNSUPPORTED:
            case SANE_STATUS_CANCELLED:
            case SANE_STATUS_EOF:
                break;
            case SANE_STATUS_NO_DOCS:
                emit scanDone(SaneWidget::NoError, m_scanThread->saneStrStatus());
                QMessageBox::information(q, tr("Scan done"), m_scanThread->saneStrStatus());
                break;
            case SANE_STATUS_INVAL:
            case SANE_STATUS_IO_ERROR:
            case SANE_STATUS_NO_MEM:
            case SANE_STATUS_JAMMED:
            case SANE_STATUS_COVER_OPEN:
            case SANE_STATUS_DEVICE_BUSY:
            case SANE_STATUS_ACCESS_DENIED:
                emit scanDone(SaneWidget::ErrorGeneral, m_scanThread->saneStrStatus());
                QMessageBox::warning(q, tr("Sorry"), m_scanThread->saneStrStatus());
                if (m_scanThread->saneStatus() == SANE_STATUS_INVAL)
                    m_optsWidget->reloadOptions();
                break;
        }
    }
    
    sane_cancel(m_saneHandle);

    // clear the highlight
    m_previewViewer->setHighlightArea(0,0,1,1);
    setBusy(false);
    m_scanOngoing = false;
    m_cancelNextScan = false;
}

void SaneWidgetPrivate::setBusy(bool busy)
{
    if (busy) {
        m_warmingUp->show();
        m_activityFrame->hide();
        m_btnFrame->hide();
    }
    else {
        m_warmingUp->hide();
        m_activityFrame->hide();
        m_btnFrame->show();
    }
    
    m_optsWidget->setDisabled(busy);
    m_previewViewer->setDisabled(busy);
    
    m_scanBtn->setFocus(Qt::OtherFocusReason);
}

void SaneWidgetPrivate::invertPreview()
{
    m_previewImg.invertPixels();
    m_previewViewer->updateImage();
}

void SaneWidgetPrivate::scanSourceChanged()
{
    m_previewViewer->setMultipleSelectionsAllowed(!m_optsWidget->isADFSource());
}

void SaneWidgetPrivate::updateProgress()
{
    int progress;
    if (m_isPreview) {
        progress = m_previewThread->scanProgress();
        if (m_previewThread->saneStartDone()) {
            if (!m_progressBar->isVisible()) {
                m_warmingUp->hide();
                m_activityFrame->show();
                // the image size might have changed
                m_previewViewer->setQImage(&m_previewImg);
                //m_previewViewer->zoom2Fit();
            }
            else {
                m_previewViewer->updateImage();
            }
        }
    }
    else {
        progress = m_scanThread->scanProgress();
        if (m_scanThread->saneStartDone()) {
            if (!m_progressBar->isVisible())
            {
                m_warmingUp->hide();
                m_activityFrame->show();
            }
            else
                m_previewViewer->updateImage();
        }
        //m_previewViewer->setHighlightShown(progress);
    }

    m_progressBar->setValue(progress);
    emit scanProgress(progress);
}

}  // NameSpace LibQSane
